@@ -1,8 +1,10 @@
-from bw2data import Database, methods, databases, mapping, Method, config
-from time import time
+import sys
+from warnings import warn
+
 import bw2calc as bc
 import numpy as np
 import pyprind
+from bw2data import Database, databases, get_activity, methods
 
 
 def contribution_for_all_datasets_one_method(database, method, progress=True):
@@ -15,7 +17,6 @@ def contribution_for_all_datasets_one_method(database, method, progress=True):
     Returns:
         NumPy array of relative contributions. Each column sums to one.
         Lookup dictionary, dataset keys to row/column indices
-        Total elapsed time in seconds
 
     """
 
@@ -33,19 +34,17 @@ def contribution_for_all_datasets_one_method(database, method, progress=True):
         else:
             return scores / summed
 
-    start = time()
-    assert database in databases, "Can't find database %s" % database
-    assert method in methods, "Can't find method %s" % method
-    keys = Database(database).load().keys()
-    assert keys, "Database %s appears to have no datasets" % database
+    assert database in databases, f"Can't find database {database}"
+    assert method in methods, f"Can't find method {method}"
+    db = Database(database)
+    assert len(db), f"Database {database} appears to have no datasets"
 
     # Array to store results
-    results = np.zeros((len(keys), len(keys)), dtype=np.float32)
+    results = np.zeros((len(db), len(db)), dtype=np.float32)
 
     # Instantiate LCA object
-    lca = bc.LCA({keys[0]: 1}, method=method)
+    lca = bc.LCA({db.random(): 1}, method=method)
     lca.lci()
-    lca.decompose_technosphere()
     lca.lcia()
 
     rows = lca.characterized_inventory.shape[0]
@@ -58,15 +57,15 @@ def contribution_for_all_datasets_one_method(database, method, progress=True):
         "all": np.zeros((all_cutoff, cols), dtype=np.float32),
     }
 
-    pbar = pyprind.ProgBar(len(keys), title="Activities:")
+    pbar = pyprind.ProgBar(len(db), title="Activities:")
 
     # Actual calculations
-    for key in keys:
-        lca.redo_lcia({key: 1})
-        if lca.score == 0.0:
+    for ds in db:
+        lca.redo_lcia({ds.id: 1})
+        if not lca.score:
             continue
 
-        col = lca.activity_dict[mapping[key]]
+        col = lca.dicts.activity[ds.id]
         results["activities"][:, col] = get_normalized_scores(lca, "activities")
         results["flows"][:, col] = get_normalized_scores(lca, "flows")
         results_all = get_normalized_scores(lca, "all")
@@ -80,50 +79,7 @@ def contribution_for_all_datasets_one_method(database, method, progress=True):
 
     print(pbar)
 
-    return results, lca.activity_dict, time() - start
-
-
-def group_by_emissions(method):
-    """Group characterization factors by name, realm, and unit.
-
-    **realm** is the general category, e.g. air, soil, water.
-
-    Does not work on regionalized LCIA methods!
-
-    Args:
-        *method* (tuple or Method): LCIA method
-
-    Returns:
-        Dictionary: {(name, realm, unit)}: [cfs... ]
-
-    """
-    if isinstance(method, Method):
-        data = method.load()
-    elif isinstance(method, tuple):
-        data = Method(method).load()
-    else:
-        raise ValueError("Can't interpret %s as a LCIA method" % method)
-
-    biosphere = Database(config.biosphere).load()
-    grouped = {}
-
-    for key, cf, geo in data:
-        if geo != config.global_location:
-            raise ValueError(
-                "`group_by_emissions` doesn't work on regionalized methods"
-            )
-        if key[0] != config.biosphere:
-            # Alternative biosphere, e.g. Ecoinvent 3. Add new biosphere DB
-            biosphere.update(Database(key[0]).load())
-        flow = biosphere[key]
-        label = (
-            flow.get("name", "Unknown"),
-            flow.get("categories", [""])[0],
-            flow.get("unit", "Unknown"),
-        )
-        grouped.setdefault(label, []).append(cf)
-
-    return grouped
+    return results
 
 
 def print_recursive_calculation(
@@ -162,6 +118,9 @@ def print_recursive_calculation(
         Nothing. Prints to ``sys.stdout`` or ``file_obj``
 
     """
+    activity = get_activity(activity)
+    if file_obj is None:
+        file_obj = sys.stdout
 
     if lca_obj is None:
         lca_obj = bc.LCA({activity: amount}, lcia_method)
@@ -171,15 +130,11 @@ def print_recursive_calculation(
     elif total_score is None:
         raise ValueError
     else:
-        lca_obj.redo_lcia({activity: amount})
+        lca_obj.redo_lcia({activity.id: amount})
         if abs(lca_obj.score) <= abs(total_score * cutoff):
             return
     if first:
-        message = "Fraction of score | Absolute score | Amount | Activity"
-        if file_obj is not None:
-            file_obj.write(message + "\n")
-        else:
-            print(message)
+        file_obj.write("Fraction of score | Absolute score | Amount | Activity\n")
     message = "{}{:04.3g} | {:5.4n} | {:5.4n} | {:.70}".format(
         tab_character * level,
         lca_obj.score / total_score,
@@ -187,16 +142,27 @@ def print_recursive_calculation(
         float(amount),
         str(activity),
     )
-    if file_obj is not None:
-        file_obj.write(message + "\n")
-    else:
-        print(message)
+    file_obj.write(message + "\n")
     if level < max_level:
+        prod_exchanges = list(activity.production())
+        if not prod_exchanges:
+            prod_amount = 1
+        elif len(prod_exchanges) > 1:
+            warn("Hit multiple production exchanges; aborting in this branch")
+            return
+        else:
+            prod_amount = lca_obj.technosphere_matrix[
+                lca_obj.dicts.product[prod_exchanges[0].input.id],
+                lca_obj.dicts.activity[prod_exchanges[0].output.id],
+            ]
+
         for exc in activity.technosphere():
+            if exc.input.id == exc.output.id:
+                continue
             print_recursive_calculation(
                 activity=exc.input,
                 lcia_method=lcia_method,
-                amount=amount * exc["amount"],
+                amount=amount * exc["amount"] / prod_amount,
                 max_level=max_level,
                 cutoff=cutoff,
                 first=False,
@@ -221,6 +187,8 @@ def print_recursive_supply_chain(
 
     This function is only for exploration; use ``bw2calc.GraphTraversal`` for a better performing function.
 
+    The results displayed here can also be incorrect if
+
     Args:
         activity: ``Activity``. The starting point of the supply chain graph.
         amount: int. Supply chain inputs will be scaled to this value.
@@ -234,19 +202,33 @@ def print_recursive_supply_chain(
         Nothing. Prints to ``stdout`` or ``file_obj``
 
     """
+    activity = get_activity(activity)
+    if file_obj is None:
+        file_obj = sys.stdout
 
     if cutoff > 0 and amount < cutoff:
         return
     message = "{}{:.3g}: {:.70}".format(tab_character * level, amount, str(activity))
-    if file_obj is not None:
-        file_obj.write(message + "\n")
-    else:
-        print(message)
+    file_obj.write(message + "\n")
     if level < max_level:
+        prod_exchanges = list(activity.production())
+        if not prod_exchanges:
+            prod_amount = 1
+        elif len(prod_exchanges) > 1:
+            warn("Hit multiple production exchanges; aborting in this branch")
+            return
+        else:
+            prod_amount = prod_exchanges[0]["amount"]
+            for other in activity.technosphere():
+                if other.input == prod_exchanges[0].input:
+                    prod_amount -= other["amount"]
+
         for exc in activity.technosphere():
+            if exc.input.id == exc.output.id:
+                continue
             print_recursive_supply_chain(
                 activity=exc.input,
-                amount=amount * exc["amount"],
+                amount=amount * exc["amount"] / prod_amount,
                 level=level + 1,
                 max_level=max_level,
                 cutoff=cutoff,
